@@ -1,16 +1,20 @@
 from __future__ import annotations
 
-from random import random
+from random import random, randrange
 
 from beartype import beartype
-from beartype.typing import Callable
+from beartype.typing import Callable, Literal
 
-from functools import partial
+from functools import partial, wraps
+from itertools import count
 
 import torch
 import torch.nn.functional as F
-from torch import pi, nn, tensor, is_tensor
+from torch import pi, nn, stack, tensor, is_tensor
 from torch.nn import Module, ModuleList
+from torch.distributions import Normal
+from torch.distributions.beta import Beta
+
 from torch.utils._pytree import tree_map, tree_flatten, tree_unflatten
 
 from torchdiffeq import odeint
@@ -26,10 +30,12 @@ from rotary_embedding_torch import (
 
 import einx
 from einops.layers.torch import Rearrange
-from einops import rearrange, repeat, einsum, pack, unpack
+from einops import rearrange, repeat, reduce, einsum, pack, unpack
 
 from pi_zero_pytorch.tensor_typing import Float, Int, Bool
 
+from hyper_connections import HyperConnections
+
 import tqdm
 
 # ein notation
@@ -48,6 +54,7 @@ import tqdm
 # h - image height
 # w - image width
 # f - image frames
+# s - residual streams (hyper connections paper)
 
 # token layout for transformer
 # vision and language tokens are autoregressive causal mask, actions, interal states + joint bidirectional amongst own tokens, but still autoregressive with respect to other tokens
@@ -110,6 +117,22 @@ def exists(v):
 def default(v, d):
     return v if exists(v) else d
 
+def identity(t):
+    return t
+
+def xnor(x, y):
+    return not (x ^ y)
+
+def maybe(fn):
+    @wraps(fn)
+    def inner(t, *args, **kwargs):
+        if not exists(t):
+            return None
+
+        return fn(t, *args, **kwargs)
+
+    return inner
+
 # tensor helpers
 
 def log(t, eps = 1e-20):
@@ -180,6 +203,20 @@ def pad_at_dim(
 
 # flow related
 
+def default_sample_times(
+    shape,
+    s = 0.999,
+    alpha = 1.5,
+    beta = 1,
+    device = None
+):
+    """ they propose to sample times from Beta distribution - last part of appendix part B """
+
+    alpha = torch.full(shape, alpha, device = device)
+    beta = torch.full(shape, beta, device = device)
+    sampled = Beta(alpha, beta).sample()
+    return (1. - sampled) * s
+
 def noise_assignment(data, noise):
     device = data.device
     data, noise = tuple(rearrange(t, 'b ... -> b (...)') for t in (data, noise))
@@ -187,6 +224,46 @@ def noise_assignment(data, noise):
     _, assign = linear_sum_assignment(dist.cpu())
     return torch.from_numpy(assign).to(device)
 
+# policy optimization related
+
+class MeanVarianceLoss(Module):
+    def forward(self, mu_sigma, target):
+
+        mean, variance = mu_sigma.unbind(dim = -1)
+        dist = Normal(mean, variance)
+        log_probs = dist.log_prob(target)
+
+        return -log_probs.mean()
+
+class LinearToMeanVariance(Module):
+    def __init__(
+        self,
+        dim,
+        dim_out,
+        variance_activation: Literal['softplus', 'exp'] = 'exp'
+    ):
+        super().__init__()
+
+        if variance_activation == 'softplus':
+            variance_act = F.softplus
+        elif variance_activation == 'exp':
+            variance_act = torch.exp
+        else:
+            raise ValueError(f'unrecognized activation {variance_activation}')
+
+        self.variance_act = variance_act
+
+        self.linear = nn.Sequential(
+            LinearNoBias(dim, dim_out * 2),
+            Rearrange('... (d mu_sigma) -> ... d mu_sigma', mu_sigma = 2)
+        )
+
+    def forward(self, embed):
+        out = self.linear(embed)
+        mean, variance = out.unbind(dim = -1)
+        variance = self.variance_act(variance)
+        return stack((mean, variance), dim = -1)
+
 # attention
 
 class Attention(Module):
@@ -198,7 +275,7 @@ class Attention(Module):
         heads = 8,
         dropout = 0.,
         softclamp_value = 50.,
-        num_recurrent_memory_tokens = 0,
+        accept_memories = False,
         learned_value_action_residual_mix = False,
         rotary_emb: RotaryEmbedding | None = None
     ):
@@ -218,6 +295,14 @@ class Attention(Module):
         self.to_qkv = LinearNoBias(dim, 3 * dim_inner)
         self.to_out = LinearNoBias(dim_inner, dim)
 
+        # maybe memory parameters
+
+        self.accept_memories = accept_memories
+
+        self.mem_rmsnorm = nn.RMSNorm(dim) if accept_memories else None
+        self.to_mem_qkv = LinearNoBias(dim, 3 * dim_inner) if accept_memories else None
+        self.to_mem_out = LinearNoBias(dim_inner, dim) if accept_memories else None
+
         # action parameters
 
         self.to_actions_qkvg = LinearNoBias(dim, 4 * dim_inner)
@@ -232,22 +317,11 @@ class Attention(Module):
 
         self.softclamp_value = softclamp_value
 
-        # maybe recurrent memory parameters
-
-        has_recurrent_memories = num_recurrent_memory_tokens > 0
-        self.accepts_recurrent_memories = has_recurrent_memories
-        self.num_mem = num_recurrent_memory_tokens
-
-        if not has_recurrent_memories:
-            return
-
-        self.to_memories_qkv = LinearNoBias(dim, 3 * dim_inner)
-        self.to_memories_out = LinearNoBias(dim_inner, dim)
-
     def forward_actions_with_cached_state(
         self,
         actions,
         cached_state_keys_values: tuple[Tensor, Tensor],
+        memories: tuple[Tensor, Tensor] | None = None,
         rotary_emb = None,
         mask: Bool['b n'] | None = None,
         actions_value_residual: Tensor | None = None,
@@ -265,8 +339,19 @@ class Attention(Module):
         q = aq
         mk, mv = cached_state_keys_values
 
+        # concat cache key / values with action key / values
+
         k, v = tuple(torch.cat(tensors, dim = -2) for tensors in zip((mk, mv), (ak, av)))
 
+        # handle read, write memories
+
+        assert not (self.accept_memories ^ exists(memories))
+
+        if exists(memories):
+            _, write_memories = memories
+            write_memories = self.mem_rmsnorm(write_memories)
+            # mqkv_write = self.to_mem_qkv(write_memories)
+
         if exists(rotary_emb):
             q = apply_rotary_emb(rotary_emb, q, freqs_seq_dim = -2)
             k = apply_rotary_emb(rotary_emb, k)
@@ -354,6 +439,7 @@ class Attention(Module):
         multimodal_seq,
         actions,
         rotary_emb = None,
+        memories: tuple[Tensor, Tensor] | None = None,
         mask: Bool['b n'] | None = None,
         actions_value_residual: Tensor | None = None,
         return_keys_values = False,
@@ -377,6 +463,26 @@ class Attention(Module):
 
         q, k, v = tuple(torch.cat(tensors, dim = -2) for tensors in zip((mq, mk, mv), (aq, ak, av)))
 
+        # handle read, write memories
+
+        has_memories = exists(memories) and any([m.numel() > 0 for m in memories])
+
+        assert not (self.accept_memories ^ has_memories)
+
+        if has_memories:
+            memories, unpack_memories = pack_with_inverse(memories, 'b * d')
+            memories = self.mem_rmsnorm(memories)
+            mqkv = self.to_mem_qkv(memories)
+            mqkv_read, mqkv_write = unpack_memories(mqkv, 'b * d')
+
+            mqr, mkr, mvr, mqw, mkw, mvw = tuple(self.split_heads(t) for t in (*mqkv_read.chunk(3, dim = -1), *mqkv_write.chunk(3, dim = -1)))
+
+            k = torch.cat((mkr, k, mkw), dim = -2)
+            v = torch.cat((mvr, v, mvw), dim = -2)
+            q, attn_output_unpack_memories = pack_with_inverse((mqr, q, mqw), 'b h * d')
+
+        # rotary embedding
+
         if exists(rotary_emb):
             q = apply_rotary_emb(rotary_emb, q)
             k = apply_rotary_emb(rotary_emb, k)
@@ -415,6 +521,11 @@ class Attention(Module):
 
         out = out * gates
 
+        # split out memories
+
+        if self.accept_memories:
+            mem_read_out, out, mem_write_out = attn_output_unpack_memories(out)
+
         # merge attention heads
 
         out = self.merge_heads(out)
@@ -425,6 +536,13 @@ class Attention(Module):
 
         output =  self.to_out(mout), self.to_actions_out(aout)
 
+        if self.accept_memories:
+            mem_out, unpack_memories = pack_with_inverse((mem_read_out, mem_write_out), 'b h * d')
+            mem_out = self.merge_heads(mem_out)
+            mem_out = self.to_mem_out(mem_out)
+
+            output = (*output, unpack_memories(mem_out, 'b * d'))
+
         if not return_keys_values:
             return output
 
@@ -550,8 +668,13 @@ class PiZero(Module):
         lm_loss_weight = 1.,
         flow_loss_weight = 1.,
         immiscible_flow = False, # https://arxiv.org/abs/2406.12303
+        sample_times_fn = default_sample_times,
         reward_tokens_dropout_prob = 0.,
         num_recurrent_memory_tokens = 0,
+        num_residual_streams = 1,
+        dim_latent_gene = None,
+        policy_optimizable = False, # if set to True, will use mean variance network for access to log prob
+        flow_variance_activation: Literal['exp', 'softplus'] = 'exp',
         odeint_kwargs: dict = dict(
             atol = 1e-5,
             rtol = 1e-5,
@@ -608,6 +731,16 @@ class PiZero(Module):
             nn.SiLU(),
         )
 
+        # latent variable / gene conditioning
+
+        needs_latent = exists(dim_latent_gene)
+        self.needs_latent = needs_latent
+
+        self.to_latent_cond = nn.Sequential(
+            nn.Linear(dim_latent_gene, dim_time_cond),
+            nn.SiLU()
+        ) if needs_latent else None
+
         # positional embedding
 
         self.rotary_emb = RotaryEmbedding(dim_head)
@@ -621,6 +754,14 @@ class PiZero(Module):
 
         self.final_norm_write_memories = nn.RMSNorm(dim) if self.has_recurrent_memories else None
 
+        # residual functions, with maybe hyper connections
+
+        assert num_residual_streams >= 1
+        init_residual_fn, self.maybe_expand_residuals, self.maybe_reduce_residuals = HyperConnections.get_init_and_expand_reduce_stream_functions(num_residual_streams, disable = num_residual_streams == 1)
+
+        residual_fns = []
+        counter = count()
+
         # attention and feedforward
 
         layers = []
@@ -630,12 +771,17 @@ class PiZero(Module):
             is_first_block = i == 0
 
             layers.append(ModuleList([
-                Attention(dim = dim, dim_head = dim_head, heads = heads, num_recurrent_memory_tokens = num_recurrent_memory_tokens, learned_value_action_residual_mix = not is_first_block, **attn_kwargs),
+                Attention(dim = dim, dim_head = dim_head, heads = heads, accept_memories = self.has_recurrent_memories, learned_value_action_residual_mix = not is_first_block, **attn_kwargs),
                 SwiGLUFeedForward(dim = dim, expand_factor = ff_expand_factor, **ff_kwargs),
                 SwiGLUFeedForward(dim = dim, expand_factor = ff_expand_factor, rmsnorm = False, **ff_kwargs),
                 SwiGLUFeedForward(dim = dim, expand_factor = ff_expand_factor, **ff_kwargs) if self.has_recurrent_memories else None
             ]))
 
+            residual_fns.append(ModuleList([
+                init_residual_fn(dim = dim, layer_index = next(counter)),
+                init_residual_fn(dim = dim, layer_index = next(counter)),
+            ]))
+
             cond_layers.append(ModuleList([
                 AdaptiveRMSNorm(dim, dim_time_cond),
                 AdaptiveLayerscale(dim, dim_time_cond),
@@ -646,6 +792,8 @@ class PiZero(Module):
         self.layers = ModuleList(layers)
         self.cond_layers = ModuleList(cond_layers)
 
+        self.residual_layers = ModuleList(residual_fns)
+
         self.final_norm_softclamp = partial(softclamp, value = final_norm_softclamp_value)
 
         self.final_norm = nn.RMSNorm(dim)
@@ -654,7 +802,16 @@ class PiZero(Module):
         # unembedding
 
         self.state_to_logits = LinearNoBias(dim, num_tokens)
-        self.actions_to_pred_flow = LinearNoBias(dim, dim_action_input)
+
+        if not policy_optimizable:
+            self.actions_to_pred_flow = LinearNoBias(dim, dim_action_input)
+            self.loss_fn = nn.MSELoss()
+        else:
+            self.actions_to_pred_flow = LinearToMeanVariance(dim, dim_action_input, variance_activation = flow_variance_activation)
+            self.loss_fn = MeanVarianceLoss()
+
+        self.is_mean_variance_output = policy_optimizable
+        self.policy_optimizable = policy_optimizable
 
         # the language token id padding id, for fine-tuning as well as taking care of the masking on top of causal mask
 
@@ -668,6 +825,10 @@ class PiZero(Module):
 
         self.reward_tokens_dropout_prob = reward_tokens_dropout_prob
 
+        # time sampling related
+
+        self.sample_times_fn = default(sample_times_fn, torch.rand)
+
         # loss related
 
         self.lm_loss_weight = lm_loss_weight
@@ -716,21 +877,25 @@ class PiZero(Module):
 
         return ema_pi_zero
 
-    @torch.inference_mode()
+    @torch.no_grad()
     def sample_actions(
         self,
         images,
         token_ids,
         joint_states,
         trajectory_length: int,
-        reward_tokens: Float['b d'] = None,
+        latents: Float['d'] | Float['b d'] = None,
+        reward_tokens: Float['b d'] | None = None,
         internal_state_tokens: Float['b ns d'] | None = None,
         steps = 18,
         show_pbar = True,
         cond_scale = 0.,
+        temperature = 1.,
         remove_parallel_component = True,
         keep_parallel_frac = 0.,
-        cache_kv = True
+        cache_kv = True,
+        return_states_for_replay = False,
+        critic: Module | None = None,
     ):
         batch_size = token_ids.shape[0]
 
@@ -739,6 +904,17 @@ class PiZero(Module):
 
         pbar = tqdm.tqdm(desc = 'sampling action trajectory', disable = not show_pbar, total = steps)
 
+        # accumulate log probs for ppo
+
+        assert not (return_states_for_replay and not self.is_mean_variance_output), 'only pi-zero with `policy_optimizable` turned on can return log probs'
+
+        timesteps = []
+        log_probs = []
+        sampled_flows = []
+        denoised_actions_across_time = []
+
+        critic_values = []
+
         # ode step function
 
         cached_state_kv = None
@@ -748,20 +924,46 @@ class PiZero(Module):
             nonlocal cached_state_kv
             nonlocal null_cached_state_kv
 
-            flow, (new_cached_state_kv, new_null_cached_state_kv) = self.forward_with_reward_cfg(
+            input_args = (
                 images,
                 token_ids,
                 joint_states,
-                denoised_actions,
+                denoised_actions
+            )
+
+            input_kwargs = dict(
                 times = timestep,
+                latents = latents,
                 reward_tokens = reward_tokens,
                 internal_state_tokens = internal_state_tokens,
                 cached_state_keys_values = (cached_state_kv, null_cached_state_kv),
                 cond_scale = cond_scale,
                 remove_parallel_component = remove_parallel_component,
-                keep_parallel_frac = keep_parallel_frac,
+                keep_parallel_frac = keep_parallel_frac
             )
 
+            output, (new_cached_state_kv, new_null_cached_state_kv) = self.forward_with_reward_cfg(*input_args, **input_kwargs)
+
+            if exists(critic):
+                critic_value, _ = critic.forward_with_reward_cfg(*input_args, **input_kwargs)
+                critic_values.append(critic_value)
+
+            flow = output
+
+            if self.is_mean_variance_output:
+                mean, variance = output.unbind(dim = -1)
+
+                flow = torch.normal(mean, variance * temperature)
+
+                log_prob = Normal(mean, variance).log_prob(flow)
+
+                # save for replaying for optimizing actor
+
+                denoised_actions_across_time.append(denoised_actions)
+                timesteps.append(repeat(timestep, ' -> b', b = batch_size))
+                log_probs.append(log_prob)
+                sampled_flows.append(flow)
+
             if cache_kv:
                 cached_state_kv = new_cached_state_kv
                 null_cached_state_kv = new_null_cached_state_kv
@@ -788,7 +990,25 @@ class PiZero(Module):
 
         pbar.close()
 
-        return sampled_actions
+        if not return_states_for_replay:
+            return sampled_actions
+
+        # place the time step dimension after batch
+
+        timesteps = stack(timesteps, dim = 1)
+        log_probs = stack(log_probs, dim = 1)
+        sampled_flow = stack(sampled_flows, dim = 1)
+        denoised_actions_across_time = stack(denoised_actions_across_time, dim = 1)
+
+        policy_optimization_outputs = (denoised_actions_across_time, timesteps, log_probs, sampled_flow)
+
+        # return critic value predictions if passed in - will just deepcopy pi-zero + a critic head
+
+        if exists(critic):
+            critic_values = stack(critic_values, dim = 1)
+            policy_optimization_outputs = (*policy_optimization_outputs, critic_values)
+
+        return sampled_actions, policy_optimization_outputs
 
     @torch.no_grad()
     def forward_with_reward_cfg(
@@ -906,6 +1126,83 @@ class PiZero(Module):
 
         return logits
 
+    def forward_for_policy_loss(
+        self,
+        images,
+        commands,
+        joint_state,
+        actions,
+        flow,
+        times,
+        old_log_probs: Float['b na'],
+        advantages: Float['b'],
+        clip_eps = 0.2,
+        entropy_weight = 1e-2,
+        norm_eps = 1e-5,
+        **kwargs,
+    ):
+
+        assert self.policy_optimizable
+        assert 'return_actions_flow' not in kwargs
+
+        # flatten the time into the batch for actions at timestep, sampled flow, and log prob
+
+        if times.ndim == 2:
+            times = rearrange(times, 'b t -> (b t)')
+
+        if flow.ndim == 4:
+            flow = rearrange(flow, 'b t ... -> (b t) ...')
+
+        if old_log_probs.ndim == 4:
+            old_log_probs = rearrange(old_log_probs, 'b t ... -> (b t) ...')
+
+        if actions.ndim == 4:
+            actions = rearrange(actions, 'b t ... -> (b t) ...')
+
+        # expand inputs across timesteps if need be
+
+        (
+            images,
+            commands,
+            joint_state,
+        ) = tuple(repeat(inp, 'b ... -> (b t) ...', t = times.shape[0] // inp.shape[0]) for inp in (
+            images,
+            commands,
+            joint_state
+        ))
+
+        mean_variance = self.forward(
+            images,
+            commands,
+            joint_state,
+            actions,
+            return_actions_flow = True,
+            **kwargs
+        )
+
+        normal_dist = Normal(*mean_variance.unbind(dim = -1))
+
+        new_log_probs = normal_dist.log_prob(actions)
+
+        # ppo surrogate loss
+
+        ratio = (new_log_probs - old_log_probs).exp()
+
+        advantages = F.layer_norm(advantages, advantages.shape, eps = norm_eps)
+
+        advantages = repeat(advantages, 'b -> (b t) 1 1', t = times.shape[0] // advantages.shape[0])
+
+        surr1 = ratio * advantages
+        surr2 = ratio.clamp(1. - clip_eps, 1. + clip_eps) * advantages
+
+        clipped_surr_loss = torch.min(surr1, surr2).sum(dim = -1)
+
+        # entropy
+
+        entropy = (normal_dist.entropy() * entropy_weight).sum(dim = -1)
+
+        return -(clipped_surr_loss + entropy * entropy_weight).mean()
+
     def forward(
         self,
         images: Float['b nv d'] | Float['b c h w'] | Float['b c f h w'], # vision
@@ -913,6 +1210,7 @@ class PiZero(Module):
         joint_state: Float['b djs'],                                     # joint state
         actions: Float['b na da'] | None = None,                         # action
         times: Float['b'] = None,
+        latents: Float['d'] | Float['b d'] = None,
         reward_tokens: Float['b d'] | None = None,
         internal_state_tokens: Float['b ns d'] | None = None,
         external_states: tuple[Float['b ...']] | None = None,
@@ -936,11 +1234,16 @@ class PiZero(Module):
         # noising the action for flow matching
 
         if not exists(times):
-            times = torch.rand((batch,), device = device)
+            times = self.sample_times_fn((batch,), device = device)
 
         if times.ndim == 0:
             times = repeat(times, '-> b', b = batch)
 
+        # handle latent genes
+
+        if exists(latents) and latents.ndim == 1:
+            latents = repeat(latents, 'd -> b d', b = batch)
+
         # if not returning the actions predicted flow, assume training and noise the actions for loss
 
         if not return_actions_flow:
@@ -960,6 +1263,15 @@ class PiZero(Module):
         time_cond = self.to_time_cond(times)
         action_tokens = self.to_action_tokens(actions)
 
+        # handle maybe latents
+
+        assert xnor(exists(latents), self.needs_latent)
+
+        if self.needs_latent:
+            latent_cond = self.to_latent_cond(latents)
+
+            time_cond = time_cond + latent_cond
+
         # register tokens
 
         action_register_tokens = repeat(self.action_register_tokens, '... -> b ...', b = batch)
@@ -992,6 +1304,8 @@ class PiZero(Module):
 
         memory_tokens = (past_recurrent_memory_tokens, write_memory_tokens)
 
+        # mem_length = past_recurrent_memory_tokens.shape[-2] + write_memory_tokens.shape[-2]
+
         # pack into [action registers] [internal + joint states] [actions]
 
         action_tokens, inverse_pack_action_registers = pack_with_inverse([
@@ -1003,6 +1317,8 @@ class PiZero(Module):
 
         action_with_registers_length = action_tokens.shape[-2]
 
+        state_tokens = None
+
         if not inferencing:
             # language
 
@@ -1077,6 +1393,10 @@ class PiZero(Module):
 
         mask = F.pad(language_mask, (state_length - command_length, action_with_registers_length), value = True) # assume fixed number of images for now, but address variable length modality states later
 
+        # memory
+
+        mask = F.pad(mask, (past_recurrent_memory_tokens.shape[-2], write_memory_tokens.shape[-2]), value = True)
+
         # rotary embeddings
 
         seq = mask.float().cumsum(dim = -1)
@@ -1100,7 +1420,8 @@ class PiZero(Module):
                 ),
                 Q_LEN = seq_len,
                 KV_LEN = seq_len,
-                device = state_tokens.device
+                device = state_tokens.device,
+                _compile = True,
             )
 
             score_mod_fn = softclamp_score_mod(self.attn_softclamp_value)
@@ -1115,46 +1436,77 @@ class PiZero(Module):
 
         cached_state_key_values_iter = iter(default(cached_state_keys_values, []))
 
-        state_cached_keys_values = []
-
         # value residual learning
 
         actions_value_residual = None
 
+        # maybe expand residual streams
+
+        action_tokens = self.maybe_expand_residuals(action_tokens)
+
         # transformer
 
         if not inferencing:
+
+            next_state_cached_keys_values = []
+
             for (
                 (attn, state_ff, actions_ff, memories_ff),
-                (attn_ada_rmsnorm, attn_ada_layerscale, ff_ada_rmsnorm, ff_ada_layerscale)
-            ) in zip(self.layers, self.cond_layers):
+                (attn_ada_rmsnorm, attn_ada_layerscale, ff_ada_rmsnorm, ff_ada_layerscale),
+                (attn_residual, actions_ff_residual),
+            ) in zip(self.layers, self.cond_layers, self.residual_layers):
+
+                # joint attention
+
+                action_tokens, add_action_residual = attn_residual(action_tokens)
 
                 action_tokens = attn_ada_rmsnorm(action_tokens, time_cond)
 
-                (state_attn_out, actions_attn_out), (state_keys, state_values, action_keys, action_values) = attn(
+                (state_attn_out, actions_attn_out, *maybe_mem_out), (state_keys, state_values, action_keys, action_values) = attn(
                     state_tokens,
                     action_tokens,
                     rotary_emb = rotary_emb,
                     flex_attn_fn = flex_attn_fn,
                     actions_value_residual = actions_value_residual,
                     mask = mask,
-                    return_keys_values = True
+                    return_keys_values = True,
+                    memories = memory_tokens
                 )
 
-                state_cached_keys_values.append((state_keys, state_values))
+                next_state_cached_keys_values.append((state_keys, state_values))
 
                 actions_value_residual = default(actions_value_residual, action_values)
 
-                state_tokens = state_tokens + state_attn_out
-                action_tokens = action_tokens + attn_ada_layerscale(actions_attn_out, time_cond)
+                action_attn_out = attn_ada_layerscale(actions_attn_out, time_cond)
 
-                state_tokens = state_ff(state_tokens) + state_tokens
+                state_tokens = state_tokens + state_attn_out
+                action_tokens = add_action_residual(action_attn_out)
+
+                if self.has_recurrent_memories:
+                    (read_mem_attn_out, write_mem_attn_out), = maybe_mem_out
+                    read_mem, write_mem = memory_tokens
+
+                    memory_tokens = (read_mem + read_mem_attn_out, write_mem + write_mem_attn_out)
+
+                # state feedforward
+
+                state_tokens_out = state_ff(state_tokens)
+
+                state_tokens = state_tokens + state_tokens_out
+
+                # action feedforward
+
+                action_tokens, add_action_ff_residual = actions_ff_residual(action_tokens)
 
                 action_tokens = ff_ada_rmsnorm(action_tokens, time_cond)
 
-                action_tokens = actions_ff(action_tokens) + action_tokens
+                action_tokens_out = actions_ff(action_tokens)
 
-                action_tokens = ff_ada_layerscale(action_tokens, time_cond)
+                action_tokens_out = ff_ada_layerscale(action_tokens_out, time_cond)
+
+                action_tokens = add_action_ff_residual(action_tokens_out)
+
+                # maybe memory feedforward
 
                 if self.has_recurrent_memories:
                     memory_tokens, unpack_memory = pack_with_inverse(memory_tokens, 'b * d')
@@ -1165,10 +1517,19 @@ class PiZero(Module):
 
         else:
 
+            assert exists(cached_state_keys_values) and len(cached_state_keys_values) > 0
+
+            next_state_cached_keys_values = cached_state_keys_values
+
             for (
                 (attn, state_ff, actions_ff, memories_ff),
-                (attn_ada_rmsnorm, attn_ada_layerscale, ff_ada_rmsnorm, ff_ada_layerscale)
-            ) in zip(self.layers, self.cond_layers):
+                (attn_ada_rmsnorm, attn_ada_layerscale, ff_ada_rmsnorm, ff_ada_layerscale),
+                (attn_residual, actions_ff_residual),
+            ) in zip(self.layers, self.cond_layers, self.residual_layers):
+
+                # actions attention
+
+                action_tokens, add_action_residual = attn_residual(action_tokens)
 
                 action_tokens = attn_ada_rmsnorm(action_tokens, time_cond)
 
@@ -1180,17 +1541,24 @@ class PiZero(Module):
                     return_keys_values = True
                 )
 
-                state_cached_keys_values.append((state_keys, state_values))
-
                 actions_value_residual = default(actions_value_residual, action_values)
 
-                action_tokens = action_tokens + attn_ada_layerscale(actions_attn_out, time_cond)
+                actions_attn_out = attn_ada_layerscale(actions_attn_out, time_cond)
+                action_tokens = add_action_residual(actions_attn_out)
+
+                # actions feed forward
+
+                action_tokens, add_action_ff_residual = actions_ff_residual(action_tokens)
 
                 action_tokens = ff_ada_rmsnorm(action_tokens, time_cond)
 
-                action_tokens = actions_ff(action_tokens) + action_tokens
+                action_out = actions_ff(action_tokens)
 
-                action_tokens = ff_ada_layerscale(action_tokens, time_cond)
+                action_out = ff_ada_layerscale(action_out, time_cond)
+
+                action_tokens = add_action_residual(action_out)
+
+                # maybe memory feed forward
 
                 if self.has_recurrent_memories:
                     memory_tokens, unpack_memory = pack_with_inverse(memory_tokens, 'b * d')
@@ -1199,6 +1567,10 @@ class PiZero(Module):
 
                     memory_tokens = unpack_memory(memory_tokens)
 
+        # maybe reduce residual streams
+
+        action_tokens = self.maybe_reduce_residuals(action_tokens)
+
         if not inferencing:
             # unpack and unembed to predictions
 
@@ -1241,13 +1613,13 @@ class PiZero(Module):
             if not return_state_keys_values:
                 return pred_actions_flow, written_memory_tokens
 
-            return pred_actions_flow, state_cached_keys_values
+            return pred_actions_flow, next_state_cached_keys_values
 
         flow_loss = self.zero
 
         if return_action_flow_loss:
-            flow_loss = F.mse_loss(flow, pred_actions_flow)
-        
+            flow_loss = self.loss_fn(pred_actions_flow, flow)
+
         # language cross entropy loss
 
         language_loss = self.zero
